# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from . import metrics_py
from ._initialize_cpp import Editops
import sys

from rapidfuzz cimport (
    RF_SCORER_FLAG_MULTI_STRING_CALL,
    RF_SCORER_FLAG_MULTI_STRING_INIT,
    RF_SCORER_FLAG_RESULT_F64,
    RF_SCORER_FLAG_RESULT_SIZE_T,
    RF_SCORER_FLAG_SYMMETRIC,
    RF_SCORER_NONE_IS_WORST_SCORE,
    RF_Kwargs,
    RF_Preprocess,
    RF_Scorer,
    RF_ScorerFlags,
    RF_ScorerFunc,
    RF_String,
    RF_UncachedScorerFunc
)

from ._initialize_cpp cimport Editops, RfEditops

pandas_NA = None

cdef inline void setupPandas() noexcept:
    global pandas_NA
    if pandas_NA is None:
        pandas = sys.modules.get('pandas')
        if hasattr(pandas, 'NA'):
            pandas_NA = pandas.NA

setupPandas()

# required for preprocess_strings
from array import array

from cpp_common cimport (
    CreateScorerContext,
    NoKwargsInit,
    RF_StringWrapper,
    SetFuncAttrs,
    SetScorerAttrs,
    preprocess_strings,
    get_score_cutoff_f64,
    get_score_cutoff_size_t,
    is_none
)
from libcpp.cmath cimport isnan
from libc.stdint cimport SIZE_MAX, int64_t
from libc.stdlib cimport free, malloc
from libcpp cimport bool
from cython.operator cimport dereference


cdef extern from "rapidfuzz/details/types.hpp" namespace "rapidfuzz" nogil:
    cdef struct LevenshteinWeightTable:
        size_t insert_cost
        size_t delete_cost
        size_t replace_cost

cdef extern from "metrics.hpp":
    # Levenshtein
    double levenshtein_normalized_distance_func(  const RF_String&, const RF_String&, size_t, size_t, size_t, double, double) except + nogil
    size_t levenshtein_distance_func(            const RF_String&, const RF_String&, size_t, size_t, size_t, size_t, size_t) except + nogil
    double levenshtein_normalized_similarity_func(const RF_String&, const RF_String&, size_t, size_t, size_t, double, double) except + nogil
    size_t levenshtein_similarity_func(          const RF_String&, const RF_String&, size_t, size_t, size_t, size_t, size_t) except + nogil

    RfEditops levenshtein_editops_func(const RF_String&, const RF_String&, size_t) except + nogil

    bool LevenshteinDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LevenshteinNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LevenshteinSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LevenshteinNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedLevenshteinDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedLevenshteinNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedLevenshteinSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedLevenshteinNormalizedSimilarityFuncInit() nogil

    bool LevenshteinMultiStringSupport(const RF_Kwargs*) nogil

    # Damerau Levenshtein
    double damerau_levenshtein_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t damerau_levenshtein_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double damerau_levenshtein_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t damerau_levenshtein_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    bool DamerauLevenshteinDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool DamerauLevenshteinNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool DamerauLevenshteinSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool DamerauLevenshteinNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedDamerauLevenshteinDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedDamerauLevenshteinNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedDamerauLevenshteinSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedDamerauLevenshteinNormalizedSimilarityFuncInit() nogil

    # LCS
    double lcs_seq_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t lcs_seq_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double lcs_seq_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t lcs_seq_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    RfEditops lcs_seq_editops_func(const RF_String&, const RF_String&) except + nogil

    bool LCSseqDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LCSseqNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LCSseqSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool LCSseqNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedLCSseqDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedLCSseqNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedLCSseqSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedLCSseqNormalizedSimilarityFuncInit() nogil

    bool LCSseqMultiStringSupport(const RF_Kwargs*) nogil

    # Indel
    double indel_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t indel_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double indel_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t indel_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    RfEditops indel_editops_func(const RF_String&, const RF_String&) except + nogil

    bool IndelDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool IndelNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool IndelSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool IndelNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedIndelDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedIndelNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedIndelSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedIndelNormalizedSimilarityFuncInit() nogil

    bool IndelMultiStringSupport(const RF_Kwargs*) nogil

    # Hamming
    double hamming_normalized_distance_func(  const RF_String&, const RF_String&, bool, double) except + nogil
    size_t hamming_distance_func(            const RF_String&, const RF_String&, bool, size_t) except + nogil
    double hamming_normalized_similarity_func(const RF_String&, const RF_String&, bool, double) except + nogil
    size_t hamming_similarity_func(          const RF_String&, const RF_String&, bool, size_t) except + nogil

    bool HammingDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool HammingNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool HammingSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool HammingNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedHammingDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedHammingNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedHammingSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedHammingNormalizedSimilarityFuncInit() nogil

    RfEditops hamming_editops_func(const RF_String&, const RF_String&, bool) except + nogil

    # Optimal String Alignment
    double osa_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t osa_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double osa_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t osa_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    bool OSADistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool OSANormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool OSASimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool OSANormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedOSADistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedOSANormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedOSASimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedOSANormalizedSimilarityFuncInit() nogil

    bool OSAMultiStringSupport(const RF_Kwargs*) nogil

    # Jaro
    double jaro_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    double jaro_distance_func(             const RF_String&, const RF_String&, double) except + nogil
    double jaro_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    double jaro_similarity_func(           const RF_String&, const RF_String&, double) except + nogil

    bool JaroDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedJaroDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedJaroNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedJaroSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedJaroNormalizedSimilarityFuncInit() nogil

    bool JaroMultiStringSupport(const RF_Kwargs*) nogil

    # Jaro Winkler
    double jaro_winkler_normalized_distance_func(  const RF_String&, const RF_String&, double, double) except + nogil
    double jaro_winkler_distance_func(             const RF_String&, const RF_String&, double, double) except + nogil
    double jaro_winkler_normalized_similarity_func(const RF_String&, const RF_String&, double, double) except + nogil
    double jaro_winkler_similarity_func(           const RF_String&, const RF_String&, double, double) except + nogil

    bool JaroWinklerDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroWinklerNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroWinklerSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool JaroWinklerNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedJaroWinklerDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedJaroWinklerNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedJaroWinklerSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedJaroWinklerNormalizedSimilarityFuncInit() nogil

    bool JaroWinklerMultiStringSupport(const RF_Kwargs*) nogil

    # Prefix
    double prefix_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t prefix_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double prefix_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t prefix_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    bool PrefixDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PrefixNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PrefixSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PrefixNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedPrefixDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedPrefixNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedPrefixSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedPrefixNormalizedSimilarityFuncInit() nogil

    # Postfix
    double postfix_normalized_distance_func(  const RF_String&, const RF_String&, double) except + nogil
    size_t postfix_distance_func(            const RF_String&, const RF_String&, size_t) except + nogil
    double postfix_normalized_similarity_func(const RF_String&, const RF_String&, double) except + nogil
    size_t postfix_similarity_func(          const RF_String&, const RF_String&, size_t) except + nogil

    bool PostfixDistanceInit(            RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PostfixNormalizedDistanceInit(  RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PostfixSimilarityInit(          RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    bool PostfixNormalizedSimilarityInit(RF_ScorerFunc*, const RF_Kwargs*, int64_t, const RF_String*) except False nogil
    RF_UncachedScorerFunc UncachedPostfixDistanceFuncInit(            ) nogil
    RF_UncachedScorerFunc UncachedPostfixNormalizedDistanceFuncInit(  ) nogil
    RF_UncachedScorerFunc UncachedPostfixSimilarityFuncInit(          ) nogil
    RF_UncachedScorerFunc UncachedPostfixNormalizedSimilarityFuncInit() nogil


def levenshtein_distance(s1, s2, *, weights=(1,1,1), processor=None, score_cutoff=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef size_t insertion, deletion, substitution
    insertion = deletion = substitution = 1
    if weights is not None:
        insertion, deletion, substitution = weights

    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, SIZE_MAX, 0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return levenshtein_distance_func(s1_proc.string, s2_proc.string, insertion, deletion, substitution, c_score_cutoff, c_score_hint)


def levenshtein_similarity(s1, s2, *, weights=(1,1,1), processor=None, score_cutoff=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef size_t insertion, deletion, substitution
    insertion = deletion = substitution = 1
    if weights is not None:
        insertion, deletion, substitution = weights

    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, 0, SIZE_MAX)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return levenshtein_similarity_func(s1_proc.string, s2_proc.string, insertion, deletion, substitution, c_score_cutoff, c_score_hint)


def levenshtein_normalized_distance(s1, s2, *, weights=(1,1,1), processor=None, score_cutoff=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef size_t insertion, deletion, substitution
    insertion = deletion = substitution = 1
    if weights is not None:
        insertion, deletion, substitution = weights

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    cdef double c_score_hint = get_score_cutoff_f64(score_hint, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return levenshtein_normalized_distance_func(s1_proc.string, s2_proc.string, insertion, deletion, substitution, c_score_cutoff, c_score_hint)


def levenshtein_normalized_similarity(s1, s2, *, weights=(1,1,1), processor=None, score_cutoff=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef size_t insertion, deletion, substitution
    insertion = deletion = substitution = 1
    if weights is not None:
        insertion, deletion, substitution = weights

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    cdef double c_score_hint = get_score_cutoff_f64(score_hint, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return levenshtein_normalized_similarity_func(s1_proc.string, s2_proc.string, insertion, deletion, substitution, c_score_cutoff, c_score_hint)


def levenshtein_editops(s1, s2, *, processor=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, SIZE_MAX, 0)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = levenshtein_editops_func(s1_proc.string, s2_proc.string, c_score_hint)
    return ops


def levenshtein_opcodes(s1, s2, *, processor=None, score_hint=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, SIZE_MAX, 0)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = levenshtein_editops_func(s1_proc.string, s2_proc.string, c_score_hint)
    return ops.as_opcodes()

cdef void KwargsDeinit(RF_Kwargs* self) noexcept:
    free(<void*>self.context)

cdef bool LevenshteinKwargsInit(RF_Kwargs* self, dict kwargs) except False:
    cdef size_t insertion, deletion, substitution
    cdef LevenshteinWeightTable* weights = <LevenshteinWeightTable*>malloc(sizeof(LevenshteinWeightTable))

    if not weights:
        raise MemoryError

    insertion, deletion, substitution = kwargs.get("weights", (1, 1, 1))
    weights.insert_cost = insertion
    weights.delete_cost = deletion
    weights.replace_cost = substitution
    self.context = weights
    self.dtor = KwargsDeinit
    return True

cdef bool GetScorerFlagsLevenshteinDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    cdef LevenshteinWeightTable* weights = <LevenshteinWeightTable*>self.context
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T
    if weights.insert_cost == weights.delete_cost:
        scorer_flags.flags |= RF_SCORER_FLAG_SYMMETRIC
    if LevenshteinMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsLevenshteinSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    cdef LevenshteinWeightTable* weights = <LevenshteinWeightTable*>self.context
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T
    if weights.insert_cost == weights.delete_cost:
        scorer_flags.flags |= RF_SCORER_FLAG_SYMMETRIC
    if LevenshteinMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsLevenshteinNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    cdef LevenshteinWeightTable* weights = <LevenshteinWeightTable*>self.context
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_NONE_IS_WORST_SCORE
    if weights.insert_cost == weights.delete_cost:
        scorer_flags.flags |= RF_SCORER_FLAG_SYMMETRIC
    if LevenshteinMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsLevenshteinNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    cdef LevenshteinWeightTable* weights = <LevenshteinWeightTable*>self.context
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_NONE_IS_WORST_SCORE
    if weights.insert_cost == weights.delete_cost:
        scorer_flags.flags |= RF_SCORER_FLAG_SYMMETRIC
    if LevenshteinMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer LevenshteinDistanceContext = CreateScorerContext(LevenshteinKwargsInit, GetScorerFlagsLevenshteinDistance, LevenshteinDistanceInit, UncachedLevenshteinDistanceFuncInit())
SetScorerAttrs(levenshtein_distance, metrics_py.levenshtein_distance, &LevenshteinDistanceContext)

cdef RF_Scorer LevenshteinSimilarityContext = CreateScorerContext(LevenshteinKwargsInit, GetScorerFlagsLevenshteinSimilarity, LevenshteinSimilarityInit, UncachedLevenshteinSimilarityFuncInit())
SetScorerAttrs(levenshtein_similarity, metrics_py.levenshtein_similarity, &LevenshteinSimilarityContext)

cdef RF_Scorer LevenshteinNormalizedDistanceContext = CreateScorerContext(LevenshteinKwargsInit, GetScorerFlagsLevenshteinNormalizedDistance, LevenshteinNormalizedDistanceInit, UncachedLevenshteinNormalizedDistanceFuncInit())
SetScorerAttrs(levenshtein_normalized_distance, metrics_py.levenshtein_normalized_distance, &LevenshteinNormalizedDistanceContext)

cdef RF_Scorer LevenshteinNormalizedSimilarityContext = CreateScorerContext(LevenshteinKwargsInit, GetScorerFlagsLevenshteinNormalizedSimilarity, LevenshteinNormalizedSimilarityInit, UncachedLevenshteinNormalizedSimilarityFuncInit())
SetScorerAttrs(levenshtein_normalized_similarity, metrics_py.levenshtein_normalized_similarity, &LevenshteinNormalizedSimilarityContext)

SetFuncAttrs(levenshtein_editops, metrics_py.levenshtein_editops)
SetFuncAttrs(levenshtein_opcodes, metrics_py.levenshtein_opcodes)

def damerau_levenshtein_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return damerau_levenshtein_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def damerau_levenshtein_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return damerau_levenshtein_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def damerau_levenshtein_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return damerau_levenshtein_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def damerau_levenshtein_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return damerau_levenshtein_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

cdef bool GetScorerFlagsDamerauLevenshteinDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsDamerauLevenshteinNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1
    return True

cdef bool GetScorerFlagsDamerauLevenshteinSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsDamerauLevenshteinNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer DamerauLevenshteinDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsDamerauLevenshteinDistance, DamerauLevenshteinDistanceInit, UncachedDamerauLevenshteinDistanceFuncInit())
SetScorerAttrs(damerau_levenshtein_distance, metrics_py.damerau_levenshtein_distance, &DamerauLevenshteinDistanceContext)

cdef RF_Scorer DamerauLevenshteinNormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsDamerauLevenshteinNormalizedDistance, DamerauLevenshteinNormalizedDistanceInit, UncachedDamerauLevenshteinNormalizedDistanceFuncInit())
SetScorerAttrs(damerau_levenshtein_normalized_distance, metrics_py.damerau_levenshtein_normalized_distance, &DamerauLevenshteinNormalizedDistanceContext)

cdef RF_Scorer DamerauLevenshteinSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsDamerauLevenshteinSimilarity, DamerauLevenshteinSimilarityInit, UncachedDamerauLevenshteinSimilarityFuncInit())
SetScorerAttrs(damerau_levenshtein_similarity, metrics_py.damerau_levenshtein_similarity, &DamerauLevenshteinSimilarityContext)

cdef RF_Scorer DamerauLevenshteinNormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsDamerauLevenshteinNormalizedSimilarity, DamerauLevenshteinNormalizedSimilarityInit, UncachedDamerauLevenshteinNormalizedSimilarityFuncInit())
SetScorerAttrs(damerau_levenshtein_normalized_similarity, metrics_py.damerau_levenshtein_normalized_similarity, &DamerauLevenshteinNormalizedSimilarityContext)


def lcs_seq_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return lcs_seq_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def lcs_seq_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return lcs_seq_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def lcs_seq_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return lcs_seq_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def lcs_seq_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return lcs_seq_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def lcs_seq_editops(s1, s2, *, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = lcs_seq_editops_func(s1_proc.string, s2_proc.string)
    return ops


def lcs_seq_opcodes(s1, s2, *, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = lcs_seq_editops_func(s1_proc.string, s2_proc.string)
    return ops.as_opcodes()


cdef bool GetScorerFlagsLCSseqDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if LCSseqMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsLCSseqNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if LCSseqMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1
    return True

cdef bool GetScorerFlagsLCSseqSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if LCSseqMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsLCSseqNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if LCSseqMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer LCSseqDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsLCSseqDistance, LCSseqDistanceInit, UncachedLCSseqDistanceFuncInit())
SetScorerAttrs(lcs_seq_distance, metrics_py.lcs_seq_distance, &LCSseqDistanceContext)

cdef RF_Scorer LCSseqNormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsLCSseqNormalizedDistance, LCSseqNormalizedDistanceInit, UncachedLCSseqNormalizedDistanceFuncInit())
SetScorerAttrs(lcs_seq_normalized_distance, metrics_py.lcs_seq_normalized_distance, &LCSseqNormalizedDistanceContext)

cdef RF_Scorer LCSseqSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsLCSseqSimilarity, LCSseqSimilarityInit, UncachedLCSseqSimilarityFuncInit())
SetScorerAttrs(lcs_seq_similarity, metrics_py.lcs_seq_similarity, &LCSseqSimilarityContext)

cdef RF_Scorer LCSseqNormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsLCSseqNormalizedSimilarity, LCSseqNormalizedSimilarityInit, UncachedLCSseqNormalizedSimilarityFuncInit())
SetScorerAttrs(lcs_seq_normalized_similarity, metrics_py.lcs_seq_normalized_similarity, &LCSseqNormalizedSimilarityContext)

SetFuncAttrs(lcs_seq_editops, metrics_py.lcs_seq_editops)
SetFuncAttrs(lcs_seq_opcodes, metrics_py.lcs_seq_opcodes)

def indel_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return indel_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def indel_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return indel_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def indel_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return indel_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def indel_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return indel_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def indel_editops(s1, s2, *, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = indel_editops_func(s1_proc.string, s2_proc.string)
    return ops


def indel_opcodes(s1, s2, *, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = indel_editops_func(s1_proc.string, s2_proc.string)
    return ops.as_opcodes()


cdef bool GetScorerFlagsIndelDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if IndelMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True


cdef bool GetScorerFlagsIndelNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if IndelMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1
    return True

cdef bool GetScorerFlagsIndelSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if IndelMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True


cdef bool GetScorerFlagsIndelNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if IndelMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True


cdef RF_Scorer IndelDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsIndelDistance, IndelDistanceInit, UncachedIndelDistanceFuncInit())
SetScorerAttrs(indel_distance, metrics_py.indel_distance, &IndelDistanceContext)

cdef RF_Scorer IndelNormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsIndelNormalizedDistance, IndelNormalizedDistanceInit, UncachedIndelNormalizedDistanceFuncInit())
SetScorerAttrs(indel_normalized_distance, metrics_py.indel_normalized_distance, &IndelNormalizedDistanceContext)

cdef RF_Scorer IndelSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsIndelSimilarity, IndelSimilarityInit, UncachedIndelSimilarityFuncInit())
SetScorerAttrs(indel_similarity, metrics_py.indel_similarity, &IndelSimilarityContext)

cdef RF_Scorer IndelNormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsIndelNormalizedSimilarity, IndelNormalizedSimilarityInit, UncachedIndelNormalizedSimilarityFuncInit())
SetScorerAttrs(indel_normalized_similarity, metrics_py.indel_normalized_similarity, &IndelNormalizedSimilarityContext)

SetFuncAttrs(indel_editops, metrics_py.indel_editops)
SetFuncAttrs(indel_opcodes, metrics_py.indel_opcodes)

def hamming_distance(s1, s2, *, pad=True, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return hamming_distance_func(s1_proc.string, s2_proc.string, pad, c_score_cutoff)

def hamming_similarity(s1, s2, *, pad=True, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return hamming_similarity_func(s1_proc.string, s2_proc.string, pad, c_score_cutoff)

def hamming_normalized_distance(s1, s2, *, pad=True, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return hamming_normalized_distance_func(s1_proc.string, s2_proc.string, pad, c_score_cutoff)


def hamming_normalized_similarity(s1, s2, *, pad=True, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return hamming_normalized_similarity_func(s1_proc.string, s2_proc.string, pad, c_score_cutoff)


def hamming_editops(s1, s2, *, pad=True, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = hamming_editops_func(s1_proc.string, s2_proc.string, pad)
    return ops


def hamming_opcodes(s1, s2, *, pad=True, processor=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    cdef Editops ops = Editops.__new__(Editops)

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    ops.editops = hamming_editops_func(s1_proc.string, s2_proc.string, pad)
    return ops.as_opcodes()

cdef bool HammingKwargsInit(RF_Kwargs* self, dict kwargs) except False:
    cdef bool* pad = <bool*>malloc(sizeof(bool))

    if not pad:
        raise MemoryError

    pad[0] = <bool>kwargs.get("pad", True)
    self.context = pad
    self.dtor = KwargsDeinit
    return True

cdef bool GetScorerFlagsHammingDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsHammingNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsHammingSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsHammingNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer HammingDistanceContext = CreateScorerContext(HammingKwargsInit, GetScorerFlagsHammingDistance, HammingDistanceInit, UncachedHammingDistanceFuncInit())
SetScorerAttrs(hamming_distance, metrics_py.hamming_distance, &HammingDistanceContext)

cdef RF_Scorer HammingNormalizedDistanceContext = CreateScorerContext(HammingKwargsInit, GetScorerFlagsHammingNormalizedDistance, HammingNormalizedDistanceInit, UncachedHammingNormalizedDistanceFuncInit())
SetScorerAttrs(hamming_normalized_distance, metrics_py.hamming_normalized_distance, &HammingNormalizedDistanceContext)

cdef RF_Scorer HammingSimilarityContext = CreateScorerContext(HammingKwargsInit, GetScorerFlagsHammingSimilarity, HammingSimilarityInit, UncachedHammingSimilarityFuncInit())
SetScorerAttrs(hamming_similarity, metrics_py.hamming_similarity, &HammingSimilarityContext)

cdef RF_Scorer HammingNormalizedSimilarityContext = CreateScorerContext(HammingKwargsInit, GetScorerFlagsHammingNormalizedSimilarity, HammingNormalizedSimilarityInit, UncachedHammingNormalizedSimilarityFuncInit())
SetScorerAttrs(hamming_normalized_similarity, metrics_py.hamming_normalized_similarity, &HammingNormalizedSimilarityContext)

SetFuncAttrs(hamming_editops, metrics_py.hamming_editops)
SetFuncAttrs(hamming_opcodes, metrics_py.hamming_opcodes)

def osa_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return osa_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def osa_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return osa_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def osa_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return osa_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def osa_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return osa_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


cdef bool GetScorerFlagsOSADistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if OSAMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsOSANormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if OSAMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsOSASimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    if OSAMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsOSANormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if OSAMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer OSADistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsOSADistance, OSADistanceInit, UncachedOSADistanceFuncInit())
SetScorerAttrs(osa_distance, metrics_py.osa_distance, &OSADistanceContext)

cdef RF_Scorer OSANormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsOSANormalizedDistance, OSANormalizedDistanceInit, UncachedOSANormalizedDistanceFuncInit())
SetScorerAttrs(osa_normalized_distance, metrics_py.osa_normalized_distance, &OSANormalizedDistanceContext)

cdef RF_Scorer OSASimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsOSASimilarity, OSASimilarityInit, UncachedOSASimilarityFuncInit())
SetScorerAttrs(osa_similarity, metrics_py.osa_similarity, &OSASimilarityContext)

cdef RF_Scorer OSANormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsOSANormalizedSimilarity, OSANormalizedSimilarityInit, UncachedOSANormalizedSimilarityFuncInit())
SetScorerAttrs(osa_normalized_similarity, metrics_py.osa_normalized_similarity, &OSANormalizedSimilarityContext)

###############################################
# Jaro
###############################################

def jaro_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def jaro_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def jaro_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def jaro_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

cdef bool GetScorerFlagsJaroDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if JaroMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsJaroSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if JaroMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer JaroDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsJaroDistance, JaroDistanceInit, UncachedJaroDistanceFuncInit())
SetScorerAttrs(jaro_distance, metrics_py.jaro_distance, &JaroDistanceContext)
SetScorerAttrs(jaro_normalized_distance, metrics_py.jaro_normalized_distance, &JaroDistanceContext)

cdef RF_Scorer JaroSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsJaroSimilarity, JaroSimilarityInit, UncachedJaroSimilarityFuncInit())
SetScorerAttrs(jaro_similarity, metrics_py.jaro_similarity, &JaroSimilarityContext)
SetScorerAttrs(jaro_normalized_similarity, metrics_py.jaro_normalized_similarity, &JaroSimilarityContext)


###############################################
# JaroWinkler
###############################################

def jaro_winkler_distance(s1, s2, *, double prefix_weight=0.1, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    if prefix_weight > 1.0 or prefix_weight < 0.0:
        msg = "prefix_weight has to be in the range 0.0 - 1.0"
        raise ValueError(msg)

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_winkler_distance_func(s1_proc.string, s2_proc.string, prefix_weight, c_score_cutoff)

def jaro_winkler_similarity(s1, s2, *, double prefix_weight=0.1, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    if prefix_weight > 1.0 or prefix_weight < 0.0:
        msg = "prefix_weight has to be in the range 0.0 - 1.0"
        raise ValueError(msg)

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_winkler_similarity_func(s1_proc.string, s2_proc.string, prefix_weight, c_score_cutoff)

def jaro_winkler_normalized_distance(s1, s2, *, double prefix_weight=0.1, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    if prefix_weight > 1.0 or prefix_weight < 0.0:
        msg = "prefix_weight has to be in the range 0.0 - 1.0"
        raise ValueError(msg)

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_winkler_normalized_distance_func(s1_proc.string, s2_proc.string, prefix_weight, c_score_cutoff)

def jaro_winkler_normalized_similarity(s1, s2, *, double prefix_weight=0.1, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    if prefix_weight > 1.0 or prefix_weight < 0.0:
        msg = "prefix_weight has to be in the range 0.0 - 1.0"
        raise ValueError(msg)

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return jaro_winkler_normalized_similarity_func(s1_proc.string, s2_proc.string, prefix_weight, c_score_cutoff)

cdef bool JaroWinklerKwargsInit(RF_Kwargs * self, dict kwargs) except False:
    cdef double * prefix_weight = <double *> malloc(sizeof(double))

    if not prefix_weight:
        raise MemoryError

    prefix_weight[0] = kwargs.get("prefix_weight", 0.1)
    if prefix_weight[0] > 1.0 or prefix_weight[0] < 0.0:
        free(prefix_weight)
        msg = "prefix_weight has to be in the range 0.0 - 1.0"
        raise ValueError(msg)

    self.context = prefix_weight
    self.dtor = KwargsDeinit
    return True

cdef bool GetScorerFlagsJaroWinklerDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if JaroWinklerMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsJaroWinklerSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    if JaroWinklerMultiStringSupport(self):
        scorer_flags.flags |= RF_SCORER_FLAG_MULTI_STRING_INIT

    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer JaroWinklerDistanceContext = CreateScorerContext(JaroWinklerKwargsInit, GetScorerFlagsJaroWinklerDistance, JaroWinklerDistanceInit, UncachedJaroWinklerDistanceFuncInit())
SetScorerAttrs(jaro_winkler_distance, metrics_py.jaro_winkler_distance, &JaroWinklerDistanceContext)
SetScorerAttrs(jaro_winkler_normalized_distance, metrics_py.jaro_winkler_normalized_distance, &JaroWinklerDistanceContext)

cdef RF_Scorer JaroWinklerSimilarityContext = CreateScorerContext(JaroWinklerKwargsInit, GetScorerFlagsJaroWinklerSimilarity, JaroWinklerSimilarityInit, UncachedJaroWinklerSimilarityFuncInit())
SetScorerAttrs(jaro_winkler_similarity, metrics_py.jaro_winkler_similarity, &JaroWinklerSimilarityContext)
SetScorerAttrs(jaro_winkler_normalized_similarity, metrics_py.jaro_winkler_normalized_similarity, &JaroWinklerSimilarityContext)

###############################################
# Postfix
###############################################

def postfix_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return postfix_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def postfix_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return postfix_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def postfix_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return postfix_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def postfix_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return postfix_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

cdef bool GetScorerFlagsPostfixDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsPostfixNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsPostfixSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsPostfixNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer PostfixDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPostfixDistance, PostfixDistanceInit, UncachedPostfixDistanceFuncInit())
SetScorerAttrs(postfix_distance, metrics_py.postfix_distance, &PostfixDistanceContext)

cdef RF_Scorer PostfixNormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPostfixNormalizedDistance, PostfixNormalizedDistanceInit, UncachedPostfixNormalizedDistanceFuncInit())
SetScorerAttrs(postfix_normalized_distance, metrics_py.postfix_normalized_distance, &PostfixNormalizedDistanceContext)

cdef RF_Scorer PostfixSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPostfixSimilarity, PostfixSimilarityInit, UncachedPostfixSimilarityFuncInit())
SetScorerAttrs(postfix_similarity, metrics_py.postfix_similarity, &PostfixSimilarityContext)

cdef RF_Scorer PostfixNormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPostfixNormalizedSimilarity, PostfixNormalizedSimilarityInit, UncachedPostfixNormalizedSimilarityFuncInit())
SetScorerAttrs(postfix_normalized_similarity, metrics_py.postfix_normalized_similarity, &PostfixNormalizedSimilarityContext)


###############################################
# Prefix
###############################################

def prefix_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, SIZE_MAX, 0)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return prefix_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def prefix_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, 0, SIZE_MAX)
    cdef RF_StringWrapper s1_proc, s2_proc

    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return prefix_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)

def prefix_normalized_distance(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 1.0, 0.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return prefix_normalized_distance_func(s1_proc.string, s2_proc.string, c_score_cutoff)


def prefix_normalized_similarity(s1, s2, *, processor=None, score_cutoff=None):
    cdef RF_StringWrapper s1_proc, s2_proc
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, 0.0, 1.0)
    preprocess_strings(s1, s2, processor, &s1_proc, &s2_proc)
    return prefix_normalized_similarity_func(s1_proc.string, s2_proc.string, c_score_cutoff)


cdef bool GetScorerFlagsPrefixDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = 0
    scorer_flags.worst_score.sizet = SIZE_MAX
    return True

cdef bool GetScorerFlagsPrefixNormalizedDistance(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 0.0
    scorer_flags.worst_score.f64 = 1.0
    return True

cdef bool GetScorerFlagsPrefixSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_SIZE_T | RF_SCORER_FLAG_SYMMETRIC
    scorer_flags.optimal_score.sizet = SIZE_MAX
    scorer_flags.worst_score.sizet = 0
    return True

cdef bool GetScorerFlagsPrefixNormalizedSimilarity(const RF_Kwargs* self, RF_ScorerFlags* scorer_flags) except False nogil:
    scorer_flags.flags = RF_SCORER_FLAG_RESULT_F64 | RF_SCORER_FLAG_SYMMETRIC | RF_SCORER_NONE_IS_WORST_SCORE
    scorer_flags.optimal_score.f64 = 1.0
    scorer_flags.worst_score.f64 = 0
    return True

cdef RF_Scorer PrefixDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPrefixDistance, PrefixDistanceInit, UncachedPrefixDistanceFuncInit())
SetScorerAttrs(prefix_distance, metrics_py.prefix_distance, &PrefixDistanceContext)

cdef RF_Scorer PrefixNormalizedDistanceContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPrefixNormalizedDistance, PrefixNormalizedDistanceInit, UncachedPrefixNormalizedDistanceFuncInit())
SetScorerAttrs(prefix_normalized_distance, metrics_py.prefix_normalized_distance, &PrefixNormalizedDistanceContext)

cdef RF_Scorer PrefixSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPrefixSimilarity, PrefixSimilarityInit, UncachedPrefixSimilarityFuncInit())
SetScorerAttrs(prefix_similarity, metrics_py.prefix_similarity, &PrefixSimilarityContext)

cdef RF_Scorer PrefixNormalizedSimilarityContext = CreateScorerContext(NoKwargsInit, GetScorerFlagsPrefixNormalizedSimilarity, PrefixNormalizedSimilarityInit, UncachedPrefixNormalizedSimilarityFuncInit())
SetScorerAttrs(prefix_normalized_similarity, metrics_py.prefix_normalized_similarity, &PrefixNormalizedSimilarityContext)



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann
from __future__ import annotations

from rapidfuzz._common_py import common_affix, conv_sequences
from rapidfuzz._utils import is_none, setupPandas
from rapidfuzz.distance import Indel_py as Indel
from rapidfuzz.distance._initialize_py import Editop, Editops


def _levenshtein_maximum(s1, s2, weights):
    len1 = len(s1)
    len2 = len(s2)
    insert, delete, replace = weights

    max_dist = len1 * delete + len2 * insert

    if len1 >= len2:
        max_dist = min(max_dist, len2 * replace + (len1 - len2) * delete)
    else:
        max_dist = min(max_dist, len1 * replace + (len2 - len1) * insert)

    return max_dist


def _uniform_generic(s1, s2, weights):
    len1 = len(s1)
    insert, delete, replace = weights
    cache = list(range(0, (len1 + 1) * delete, delete))

    for ch2 in s2:
        temp = cache[0]
        cache[0] += insert
        for i in range(len1):
            x = temp
            if s1[i] != ch2:
                x = min(cache[i] + delete, cache[i + 1] + insert, temp + replace)
            temp = cache[i + 1]
            cache[i + 1] = x

    return cache[-1]


def _uniform_distance(s1, s2):
    if not s1:
        return len(s2)

    VP = (1 << len(s1)) - 1
    VN = 0
    currDist = len(s1)
    mask = 1 << (len(s1) - 1)

    block = {}
    block_get = block.get
    x = 1
    for ch1 in s1:
        block[ch1] = block_get(ch1, 0) | x
        x <<= 1

    for ch2 in s2:
        # Step 1: Computing D0
        PM_j = block_get(ch2, 0)
        X = PM_j
        D0 = (((X & VP) + VP) ^ VP) | X | VN
        # Step 2: Computing HP and HN
        HP = VN | ~(D0 | VP)
        HN = D0 & VP
        # Step 3: Computing the value D[m,j]
        currDist += (HP & mask) != 0
        currDist -= (HN & mask) != 0
        # Step 4: Computing Vp and VN
        HP = (HP << 1) | 1
        HN = HN << 1
        VP = HN | ~(D0 | HP)
        VN = HP & D0

    return currDist


def distance(
    s1,
    s2,
    *,
    weights=(1, 1, 1),
    processor=None,
    score_cutoff=None,
    score_hint=None,
):
    """
    Calculates the minimum number of insertions, deletions, and substitutions
    required to change one sequence into the other according to Levenshtein with custom
    costs for insertion, deletion and substitution

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    weights : tuple[int, int, int] or None, optional
        The weights for the three operations in the form
        (insertion, deletion, substitution). Default is (1, 1, 1),
        which gives all three operations a weight of 1.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : int, optional
        Maximum distance between s1 and s2, that is
        considered as a result. If the distance is bigger than score_cutoff,
        score_cutoff + 1 is returned instead. Default is None, which deactivates
        this behaviour.
    score_hint : int, optional
        Expected distance between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    distance : int
        distance between s1 and s2

    Raises
    ------
    ValueError
        If unsupported weights are provided a ValueError is thrown

    Examples
    --------
    Find the Levenshtein distance between two strings:

    >>> from rapidfuzz.distance import Levenshtein
    >>> Levenshtein.distance("lewenstein", "levenshtein")
    2

    Setting a maximum distance allows the implementation to select
    a more efficient implementation:

    >>> Levenshtein.distance("lewenstein", "levenshtein", score_cutoff=1)
    2

    It is possible to select different weights by passing a `weight`
    tuple.

    >>> Levenshtein.distance("lewenstein", "levenshtein", weights=(1,1,2))
    3
    """
    _ = score_hint
    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    if weights is None or weights == (1, 1, 1):
        dist = _uniform_distance(s1, s2)
    elif weights == (1, 1, 2):
        dist = Indel.distance(s1, s2)
    else:
        dist = _uniform_generic(s1, s2, weights)

    return dist if (score_cutoff is None or dist <= score_cutoff) else score_cutoff + 1


def similarity(
    s1,
    s2,
    *,
    weights=(1, 1, 1),
    processor=None,
    score_cutoff=None,
    score_hint=None,
):
    """
    Calculates the levenshtein similarity in the range [max, 0] using custom
    costs for insertion, deletion and substitution.

    This is calculated as ``max - distance``, where max is the maximal possible
    Levenshtein distance given the lengths of the sequences s1/s2 and the weights.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    weights : tuple[int, int, int] or None, optional
        The weights for the three operations in the form
        (insertion, deletion, substitution). Default is (1, 1, 1),
        which gives all three operations a weight of 1.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : int, optional
        Maximum distance between s1 and s2, that is
        considered as a result. If the similarity is smaller than score_cutoff,
        0 is returned instead. Default is None, which deactivates
        this behaviour.
    score_hint : int, optional
        Expected similarity between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    similarity : int
        similarity between s1 and s2

    Raises
    ------
    ValueError
        If unsupported weights are provided a ValueError is thrown
    """
    _ = score_hint
    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    weights = weights or (1, 1, 1)
    maximum = _levenshtein_maximum(s1, s2, weights)
    dist = distance(s1, s2, weights=weights)
    sim = maximum - dist
    return sim if (score_cutoff is None or sim >= score_cutoff) else 0


def normalized_distance(
    s1,
    s2,
    *,
    weights=(1, 1, 1),
    processor=None,
    score_cutoff=None,
    score_hint=None,
):
    """
    Calculates a normalized levenshtein distance in the range [1, 0] using custom
    costs for insertion, deletion and substitution.

    This is calculated as ``distance / max``, where max is the maximal possible
    Levenshtein distance given the lengths of the sequences s1/s2 and the weights.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    weights : tuple[int, int, int] or None, optional
        The weights for the three operations in the form
        (insertion, deletion, substitution). Default is (1, 1, 1),
        which gives all three operations a weight of 1.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For norm_dist > score_cutoff 1.0 is returned instead. Default is None,
        which deactivates this behaviour.
    score_hint : float, optional
        Expected normalized distance between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    norm_dist : float
        normalized distance between s1 and s2 as a float between 1.0 and 0.0

    Raises
    ------
    ValueError
        If unsupported weights are provided a ValueError is thrown
    """
    _ = score_hint
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 1.0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    weights = weights or (1, 1, 1)
    maximum = _levenshtein_maximum(s1, s2, weights)
    dist = distance(s1, s2, weights=weights)
    norm_dist = dist / maximum if maximum else 0
    return norm_dist if (score_cutoff is None or norm_dist <= score_cutoff) else 1


def normalized_similarity(
    s1,
    s2,
    *,
    weights=(1, 1, 1),
    processor=None,
    score_cutoff=None,
    score_hint=None,
):
    """
    Calculates a normalized levenshtein similarity in the range [0, 1] using custom
    costs for insertion, deletion and substitution.

    This is calculated as ``1 - normalized_distance``

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    weights : tuple[int, int, int] or None, optional
        The weights for the three operations in the form
        (insertion, deletion, substitution). Default is (1, 1, 1),
        which gives all three operations a weight of 1.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_cutoff : float, optional
        Optional argument for a score threshold as a float between 0 and 1.0.
        For norm_sim < score_cutoff 0 is returned instead. Default is None,
        which deactivates this behaviour.
    score_hint : int, optional
        Expected normalized similarity between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    norm_sim : float
        normalized similarity between s1 and s2 as a float between 0 and 1.0

    Raises
    ------
    ValueError
        If unsupported weights are provided a ValueError is thrown

    Examples
    --------
    Find the normalized Levenshtein similarity between two strings:

    >>> from rapidfuzz.distance import Levenshtein
    >>> Levenshtein.normalized_similarity("lewenstein", "levenshtein")
    0.81818181818181

    Setting a score_cutoff allows the implementation to select
    a more efficient implementation:

    >>> Levenshtein.normalized_similarity("lewenstein", "levenshtein", score_cutoff=0.85)
    0.0

    It is possible to select different weights by passing a `weight`
    tuple.

    >>> Levenshtein.normalized_similarity("lewenstein", "levenshtein", weights=(1,1,2))
    0.85714285714285

    When a different processor is used s1 and s2 do not have to be strings

    >>> Levenshtein.normalized_similarity(["lewenstein"], ["levenshtein"], processor=lambda s: s[0])
    0.81818181818181
    """
    _ = score_hint
    setupPandas()
    if is_none(s1) or is_none(s2):
        return 0.0

    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    weights = weights or (1, 1, 1)
    norm_dist = normalized_distance(s1, s2, weights=weights)
    norm_sim = 1.0 - norm_dist
    return norm_sim if (score_cutoff is None or norm_sim >= score_cutoff) else 0


def _matrix(s1, s2):
    if not s1:
        return (len(s2), [], [])

    VP = (1 << len(s1)) - 1
    VN = 0
    currDist = len(s1)
    mask = 1 << (len(s1) - 1)

    block = {}
    block_get = block.get
    x = 1
    for ch1 in s1:
        block[ch1] = block_get(ch1, 0) | x
        x <<= 1

    matrix_VP = []
    matrix_VN = []
    for ch2 in s2:
        # Step 1: Computing D0
        PM_j = block_get(ch2, 0)
        X = PM_j
        D0 = (((X & VP) + VP) ^ VP) | X | VN
        # Step 2: Computing HP and HN
        HP = VN | ~(D0 | VP)
        HN = D0 & VP
        # Step 3: Computing the value D[m,j]
        currDist += (HP & mask) != 0
        currDist -= (HN & mask) != 0
        # Step 4: Computing Vp and VN
        HP = (HP << 1) | 1
        HN = HN << 1
        VP = HN | ~(D0 | HP)
        VN = HP & D0

        matrix_VP.append(VP)
        matrix_VN.append(VN)

    return (currDist, matrix_VP, matrix_VN)


def editops(
    s1,
    s2,
    *,
    processor=None,
    score_hint=None,
):
    """
    Return Editops describing how to turn s1 into s2.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_hint : int, optional
        Expected distance between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    editops : Editops
        edit operations required to turn s1 into s2

    Notes
    -----
    The alignment is calculated using an algorithm of Heikki Hyyrö, which is
    described [8]_. It has a time complexity and memory usage of ``O([N/64] * M)``.

    References
    ----------
    .. [8] Hyyrö, Heikki. "A Note on Bit-Parallel Alignment Computation."
           Stringology (2004).

    Examples
    --------
    >>> from rapidfuzz.distance import Levenshtein
    >>> for tag, src_pos, dest_pos in Levenshtein.editops("qabxcd", "abycdf"):
    ...    print(("%7s s1[%d] s2[%d]" % (tag, src_pos, dest_pos)))
     delete s1[1] s2[0]
    replace s1[3] s2[2]
     insert s1[6] s2[5]
    """
    _ = score_hint
    if processor is not None:
        s1 = processor(s1)
        s2 = processor(s2)

    s1, s2 = conv_sequences(s1, s2)
    prefix_len, suffix_len = common_affix(s1, s2)
    s1 = s1[prefix_len : len(s1) - suffix_len]
    s2 = s2[prefix_len : len(s2) - suffix_len]
    dist, VP, VN = _matrix(s1, s2)

    editops = Editops([], 0, 0)
    editops._src_len = len(s1) + prefix_len + suffix_len
    editops._dest_len = len(s2) + prefix_len + suffix_len

    if dist == 0:
        return editops

    editop_list = [None] * dist
    col = len(s1)
    row = len(s2)
    while row != 0 and col != 0:
        # deletion
        if VP[row - 1] & (1 << (col - 1)):
            dist -= 1
            col -= 1
            editop_list[dist] = Editop("delete", col + prefix_len, row + prefix_len)
        else:
            row -= 1

            # insertion
            if row and (VN[row - 1] & (1 << (col - 1))):
                dist -= 1
                editop_list[dist] = Editop("insert", col + prefix_len, row + prefix_len)
            else:
                col -= 1

                # replace (Matches are not recorded)
                if s1[col] != s2[row]:
                    dist -= 1
                    editop_list[dist] = Editop("replace", col + prefix_len, row + prefix_len)

    while col != 0:
        dist -= 1
        col -= 1
        editop_list[dist] = Editop("delete", col + prefix_len, row + prefix_len)

    while row != 0:
        dist -= 1
        row -= 1
        editop_list[dist] = Editop("insert", col + prefix_len, row + prefix_len)

    editops._editops = editop_list
    return editops


def opcodes(
    s1,
    s2,
    *,
    processor=None,
    score_hint=None,
):
    """
    Return Opcodes describing how to turn s1 into s2.

    Parameters
    ----------
    s1 : Sequence[Hashable]
        First string to compare.
    s2 : Sequence[Hashable]
        Second string to compare.
    processor : callable, optional
        Optional callable that is used to preprocess the strings before
        comparing them. Default is None, which deactivates this behaviour.
    score_hint : int, optional
        Expected distance between s1 and s2. This is used to select a
        faster implementation. Default is None, which deactivates this behaviour.

    Returns
    -------
    opcodes : Opcodes
        edit operations required to turn s1 into s2

    Notes
    -----
    The alignment is calculated using an algorithm of Heikki Hyyrö, which is
    described [9]_. It has a time complexity and memory usage of ``O([N/64] * M)``.

    References
    ----------
    .. [9] Hyyrö, Heikki. "A Note on Bit-Parallel Alignment Computation."
           Stringology (2004).

    Examples
    --------
    >>> from rapidfuzz.distance import Levenshtein

    >>> a = "qabxcd"
    >>> b = "abycdf"
    >>> for tag, i1, i2, j1, j2 in Levenshtein.opcodes("qabxcd", "abycdf"):
    ...    print(("%7s a[%d:%d] (%s) b[%d:%d] (%s)" %
    ...           (tag, i1, i2, a[i1:i2], j1, j2, b[j1:j2])))
     delete a[0:1] (q) b[0:0] ()
      equal a[1:3] (ab) b[0:2] (ab)
    replace a[3:4] (x) b[2:3] (y)
      equal a[4:6] (cd) b[3:5] (cd)
     insert a[6:6] () b[5:6] (f)
    """
    return editops(s1, s2, processor=processor, score_hint=score_hint).as_opcodes()



from __future__ import annotations

import matplotlib.pyplot as plt
import pandas as pd

results = pd.read_csv("temp/result.csv")

results *= 1000 * 1000
results["x_axis"] /= 1000 * 1000


ax = results.plot(x="x_axis")

# plt.xticks(list(range(0, 64*20+1, 64)))

plt.title("Performance comparison of the \nDamerauLevenshtein similarity in different libraries")
plt.xlabel("string length [in characters]")
plt.ylabel("runtime [μs]")
ax.set_xlim(xmin=0)
# ax.set_ylim(bottom=0)
# ax.set_yscale('log')
plt.grid()
plt.show()



from __future__ import annotations

from rapidfuzz import utils_cpp, utils_py


def test_fullProcess():
    mixed_strings = [
        "Lorem Ipsum is simply dummy text of the printing and typesetting industry.",
        "C'est la vie",
        "Ça va?",
        "Cães danados",
        "¬Camarões assados",
        "a¬ሴ€耀",
        "Á",
    ]
    mixed_strings_proc = [
        "lorem ipsum is simply dummy text of the printing and typesetting industry",
        "c est la vie",
        "ça va",
        "cães danados",
        "camarões assados",
        "a ሴ 耀",
        "á",
    ]

    for string, proc_string in zip(mixed_strings, mixed_strings_proc):
        assert utils_cpp.default_process(string) == proc_string
        assert utils_py.default_process(string) == proc_string



from __future__ import annotations

import pytest

from rapidfuzz import utils_cpp, utils_py
from tests.distance.common import JaroWinkler


def test_hash_special_case():
    assert pytest.approx(JaroWinkler.similarity([0, -1], [0, -2])) == 0.666666


def test_large_prefix_weight():
    assert pytest.approx(JaroWinkler.similarity("milyarder", "milyarderlik", prefix_weight=0.5)) == 1.0
    assert pytest.approx(JaroWinkler.similarity("milyarder", "milyarderlik", prefix_weight=1.0)) == 1.0


def test_invalid_prefix_weight():
    with pytest.raises(ValueError, match="prefix_weight has to be in the range 0.0 - 1.0"):
        JaroWinkler.similarity("milyarder", "milyarderlik", prefix_weight=-0.1)

    with pytest.raises(ValueError, match="prefix_weight has to be in the range 0.0 - 1.0"):
        JaroWinkler.similarity("milyarder", "milyarderlik", prefix_weight=1.1)


def test_edge_case_lengths():
    """
    these are largely found by fuzz tests and implemented here as regression tests
    """
    assert pytest.approx(JaroWinkler.similarity("", "")) == 1.0
    assert pytest.approx(JaroWinkler.similarity("0", "0")) == 1
    assert pytest.approx(JaroWinkler.similarity("00", "00")) == 1
    assert pytest.approx(JaroWinkler.similarity("0", "00")) == 0.85

    assert pytest.approx(JaroWinkler.similarity("0" * 65, "0" * 65)) == 1
    assert pytest.approx(JaroWinkler.similarity("0" * 64, "0" * 65)) == 0.996923
    assert pytest.approx(JaroWinkler.similarity("0" * 63, "0" * 65)) == 0.993846

    s1 = "000000001"
    s2 = "0000010"
    assert pytest.approx(JaroWinkler.similarity(s1, s2)) == 0.926984

    s1 = "10000000000000000000000000000000000000000000000000000000000000020"
    s2 = "00000000000000000000000000000000000000000000000000000000000000000"
    assert pytest.approx(JaroWinkler.similarity(s1, s2)) == 0.979487

    s1 = "00000000000000100000000000000000000000010000000000000000000000000"
    s2 = "0000000000000000000000000000000000000000000000000000000000000000000000000000001"
    assert pytest.approx(JaroWinkler.similarity(s2, s1)) == 0.95334

    s1 = "00000000000000000000000000000000000000000000000000000000000000000"
    s2 = (
        "010000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000000000000"
    )
    assert pytest.approx(JaroWinkler.similarity(s2, s1)) == 0.852344


def testCaseInsensitive():
    assert (
        pytest.approx(JaroWinkler.similarity("new york mets", "new YORK mets", processor=utils_cpp.default_process))
        == 1.0
    )
    assert (
        pytest.approx(JaroWinkler.similarity("new york mets", "new YORK mets", processor=utils_py.default_process))
        == 1.0
    )



from __future__ import annotations

import random

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

import rapidfuzz.distance._initialize_cpp as distance_cpp
import rapidfuzz.distance._initialize_py as distance_py
from rapidfuzz.distance import Editops, Levenshtein, Opcodes


def test_editops_comparison():
    """
    test comparison with Editops
    """
    ops = Levenshtein.editops("aaabaaa", "abbaaabba")
    assert ops == ops
    assert not (ops != ops)  # noqa: SIM202
    assert ops == ops.copy()
    assert not (ops != ops.copy())  # noqa: SIM202


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_editops_get_index(module):
    """
    test __getitem__ with index of Editops
    """
    ops = module.Editops(
        [
            ("delete", 1, 1),
            ("replace", 2, 1),
            ("insert", 6, 5),
            ("insert", 6, 6),
            ("insert", 6, 7),
        ],
        7,
        9,
    )

    ops_list = [
        ("delete", 1, 1),
        ("replace", 2, 1),
        ("insert", 6, 5),
        ("insert", 6, 6),
        ("insert", 6, 7),
    ]

    assert ops[0] == ops_list[0]
    assert ops[1] == ops_list[1]
    assert ops[2] == ops_list[2]
    assert ops[3] == ops_list[3]
    assert ops[4] == ops_list[4]

    assert ops[-1] == ops_list[-1]
    assert ops[-2] == ops_list[-2]
    assert ops[-3] == ops_list[-3]
    assert ops[-4] == ops_list[-4]
    assert ops[-5] == ops_list[-5]

    with pytest.raises(IndexError):
        ops[5]
    with pytest.raises(IndexError):
        ops[-6]


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_editops_get_slice(module):
    """
    test __getitem__ with slice of Editops
    """
    ops = module.Editops(
        [
            ("delete", 1, 1),
            ("replace", 2, 1),
            ("insert", 6, 5),
            ("insert", 6, 6),
            ("insert", 6, 7),
        ],
        7,
        9,
    )

    ops_list = [
        ("delete", 1, 1),
        ("replace", 2, 1),
        ("insert", 6, 5),
        ("insert", 6, 6),
        ("insert", 6, 7),
    ]

    assert ops[::].as_list() == ops_list[::]
    assert ops[::] is not ops
    assert ops[1:].as_list() == ops_list[1:]
    assert ops[:3].as_list() == ops_list[:3]
    assert ops[1:3].as_list() == ops_list[1:3]
    assert ops[:-1].as_list() == ops_list[:-1]
    assert ops[-3:].as_list() == ops_list[-3:]
    assert ops[-3:-1].as_list() == ops_list[-3:-1]

    with pytest.raises(ValueError, match="slice step cannot be zero"):
        ops[::0]

    with pytest.raises(ValueError, match="step sizes below 0 lead to an invalid order of editops"):
        ops[::-1]


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_editops_del_slice(module):
    """
    test __delitem__ with slice of Editops
    """
    ops = module.Editops(
        [
            ("delete", 1, 1),
            ("replace", 2, 1),
            ("insert", 6, 5),
            ("insert", 6, 6),
            ("insert", 6, 7),
        ],
        7,
        9,
    )

    ops_list = [
        ("delete", 1, 1),
        ("replace", 2, 1),
        ("insert", 6, 5),
        ("insert", 6, 6),
        ("insert", 6, 7),
    ]

    def del_test(key):
        _ops = ops[::]
        _ops_list = ops_list[::]
        del _ops[key]
        del _ops_list[key]
        assert _ops.as_list() == _ops_list

    del_test(slice(None, 4, None))
    del_test(slice(1, None, None))
    del_test(slice(1, 4, None))
    del_test(slice(None, 4, 2))
    del_test(slice(1, None, 2))
    del_test(slice(1, 4, 2))

    del_test(slice(None, -1, None))
    del_test(slice(-4, None, None))
    del_test(slice(-4, -1, None))
    del_test(slice(None, -1, 2))
    del_test(slice(-4, None, 2))
    del_test(slice(-4, -1, 2))


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_editops_inversion(module):
    """
    test correct inversion of Editops
    """
    ops = module.Editops(
        [
            ("delete", 1, 1),
            ("replace", 2, 1),
            ("insert", 6, 5),
            ("insert", 6, 6),
            ("insert", 6, 7),
        ],
        7,
        9,
    )

    assert ops.inverse().as_list() == [
        ("insert", 1, 1),
        ("replace", 1, 2),
        ("delete", 5, 6),
        ("delete", 6, 6),
        ("delete", 7, 6),
    ]


def test_opcodes_comparison():
    """
    test comparison with Opcodes
    """
    ops = Levenshtein.opcodes("aaabaaa", "abbaaabba")
    assert ops == ops
    assert not (ops != ops)  # noqa: SIM202
    assert ops == ops.copy()
    assert not (ops != ops.copy())  # noqa: SIM202


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_opcode_py_get_index(module):
    """
    test __getitem__ with index of Opcodes
    """
    ops = module.Opcodes(
        [
            ("equal", 0, 1, 0, 1),
            ("delete", 1, 2, 1, 1),
            ("replace", 2, 3, 1, 2),
            ("equal", 3, 6, 2, 5),
            ("insert", 6, 6, 5, 8),
            ("equal", 6, 7, 8, 9),
        ],
        7,
        9,
    )

    ops_list = [
        ("equal", 0, 1, 0, 1),
        ("delete", 1, 2, 1, 1),
        ("replace", 2, 3, 1, 2),
        ("equal", 3, 6, 2, 5),
        ("insert", 6, 6, 5, 8),
        ("equal", 6, 7, 8, 9),
    ]

    assert ops[0] == ops_list[0]
    assert ops[1] == ops_list[1]
    assert ops[2] == ops_list[2]
    assert ops[3] == ops_list[3]
    assert ops[4] == ops_list[4]
    assert ops[5] == ops_list[5]

    assert ops[-1] == ops_list[-1]
    assert ops[-2] == ops_list[-2]
    assert ops[-3] == ops_list[-3]
    assert ops[-4] == ops_list[-4]
    assert ops[-5] == ops_list[-5]
    assert ops[-6] == ops_list[-6]

    with pytest.raises(IndexError):
        ops[6]
    with pytest.raises(IndexError):
        ops[-7]


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_opcode_inversion(module):
    """
    test correct inversion of Opcodes
    """
    ops = module.Opcodes(
        [
            ("equal", 0, 1, 0, 1),
            ("delete", 1, 2, 1, 1),
            ("replace", 2, 3, 1, 2),
            ("equal", 3, 6, 2, 5),
            ("insert", 6, 6, 5, 8),
            ("equal", 6, 7, 8, 9),
        ],
        7,
        9,
    )

    assert ops.inverse().as_list() == [
        ("equal", 0, 1, 0, 1),
        ("insert", 1, 1, 1, 2),
        ("replace", 1, 2, 2, 3),
        ("equal", 2, 5, 3, 6),
        ("delete", 5, 8, 6, 6),
        ("equal", 8, 9, 6, 7),
    ]


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_opcodes_empty(module):
    """
    test behavior of conversion between empty list and Editops
    """
    ops = module.Opcodes([], 0, 0)
    assert ops.as_list() == []
    assert ops.src_len == 0
    assert ops.dest_len == 0

    ops = module.Opcodes([], 0, 3)
    assert ops.as_list() == [module.Opcode(tag="equal", src_start=0, src_end=0, dest_start=0, dest_end=3)]
    assert ops.src_len == 0
    assert ops.dest_len == 3


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_editops_empty(module):
    """
    test behavior of conversion between empty list and Opcodes
    """
    ops = module.Editops([], 0, 0)
    assert ops.as_list() == []
    assert ops.src_len == 0
    assert ops.dest_len == 0

    ops = module.Editops([], 0, 3)
    assert ops.as_list() == []
    assert ops.src_len == 0
    assert ops.dest_len == 3


def test_list_initialization():
    """
    test whether list initialization works correctly
    """
    ops = Levenshtein.opcodes("aaabaaa", "abbaaabba")
    ops2 = Opcodes(ops.as_list(), ops.src_len, ops.dest_len)
    assert ops == ops2

    ops = Levenshtein.editops("aaabaaa", "abbaaabba")
    ops2 = Editops(ops.as_list(), ops.src_len, ops.dest_len)
    assert ops == ops2

    ops = Levenshtein.opcodes("aaabaaa", "abbaaabba")
    ops2 = Editops(ops.as_list(), ops.src_len, ops.dest_len)
    assert ops.as_editops() == ops2

    ops = Levenshtein.editops("aaabaaa", "abbaaabba")
    ops2 = Opcodes(ops.as_list(), ops.src_len, ops.dest_len)
    assert ops.as_opcodes() == ops2

    ops = Levenshtein.editops("skdsakldsakdlasda", "djkajkdfkdgkhdfjrmecsidjf")
    ops2 = Opcodes(ops.as_list(), ops.src_len, ops.dest_len)
    assert ops.as_opcodes() == ops2


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_merge_adjacent_blocks(module):
    """
    test whether adjacent blocks are merged
    """
    ops1 = [module.Opcode(tag="equal", src_start=0, src_end=3, dest_start=0, dest_end=3)]
    ops2 = [
        module.Opcode(tag="equal", src_start=0, src_end=1, dest_start=0, dest_end=1),
        module.Opcode(tag="equal", src_start=1, src_end=3, dest_start=1, dest_end=3),
    ]
    assert module.Opcodes(ops1, 3, 3) == module.Opcodes(ops2, 3, 3)
    assert module.Opcodes(ops2, 3, 3) == module.Opcodes(ops2, 3, 3).as_editops().as_opcodes()


@pytest.mark.parametrize("module", [distance_py, distance_cpp])
def test_empty_matching_blocks(module):
    """
    test behavior for empty matching blocks
    """
    assert module.Editops([], 0, 0).as_matching_blocks() == [module.MatchingBlock(a=0, b=0, size=0)]
    assert module.Editops([], 0, 3).as_matching_blocks() == [module.MatchingBlock(a=0, b=3, size=0)]
    assert module.Editops([], 3, 0).as_matching_blocks() == [module.MatchingBlock(a=3, b=0, size=0)]

    assert module.Opcodes([], 0, 0).as_matching_blocks() == [module.MatchingBlock(a=0, b=0, size=0)]
    assert module.Opcodes([], 0, 3).as_matching_blocks() == [module.MatchingBlock(a=0, b=3, size=0)]
    assert module.Opcodes([], 3, 0).as_matching_blocks() == [module.MatchingBlock(a=3, b=0, size=0)]


@given(s1=st.text(), s2=st.text())
@settings(max_examples=100, deadline=None)
def test_editops_reversible(s1, s2):
    """
    test whether conversions between editops and opcodes are reversible
    """
    ops = Levenshtein.editops(s1, s2)
    assert ops == ops.as_opcodes().as_editops()
    while ops:
        del ops[random.randrange(len(ops))]
        assert Opcodes(ops.as_list(), ops.src_len, ops.dest_len) == ops.as_opcodes()
        assert ops == ops.as_opcodes().as_editops()



# todo combine benchmarks of scorers into common code base
from __future__ import annotations

import timeit

import pandas as pd


def benchmark(name, func, setup, lengths, count):
    print(f"starting {name}")
    start = timeit.default_timer()
    results = []
    for length in lengths:
        test = timeit.Timer(func, setup=setup.format(length, count))
        results.append(min(test.timeit(number=1) for _ in range(7)) / count)
    stop = timeit.default_timer()
    print(f"finished {name}, Runtime: ", stop - start)
    return results


setup = """
from rapidfuzz import fuzz as rfuzz
from fuzzywuzzy import fuzz
import string
import random
random.seed(18)
characters = string.ascii_letters + string.digits + string.whitespace + string.punctuation
a      = ''.join(random.choice(characters) for _ in range({0}))
b_list = [''.join(random.choice(characters) for _ in range({0})) for _ in range({1})]
"""

lengths = list(range(1, 512, 2))
count = 1000


def scorer_benchmark(funcname):
    time_rapidfuzz = benchmark("rapidfuzz", f"[rfuzz.{funcname}(a, b) for b in b_list]", setup, lengths, count)

    time_fuzzywuzzy = benchmark("fuzzywuzzy", f"[fuzz.{funcname}(a, b) for b in b_list]", setup, lengths, count)

    results = pd.DataFrame(
        data={
            "length": lengths,
            "rapidfuzz": time_rapidfuzz,
            "fuzzywuzzy": time_fuzzywuzzy,
        }
    )

    results.to_csv(f"results/{funcname}.csv", sep=",", index=False)


scorer_benchmark("ratio")
scorer_benchmark("partial_ratio")
scorer_benchmark("token_sort_ratio")
scorer_benchmark("token_set_ratio")
scorer_benchmark("partial_token_sort_ratio")
scorer_benchmark("partial_token_set_ratio")
scorer_benchmark("WRatio")

# token_ratio is unique to RapidFuzz
time_token_ratio = benchmark(
    "token_ratio",
    "[rfuzz.token_ratio(a, b, processor=None) for b in b_list]",
    setup,
    lengths,
    count,
)

# this gets very slow, so only benchmark it for smaller values
time_token_ratio_simple = benchmark(
    "fuzzywuzzy",
    "[max(rfuzz.token_sort_ratio(a, b, processor=None), rfuzz.token_set_ratio(a, b, processor=None)) for b in b_list]",
    setup,
    lengths,
    count,
)

results = pd.DataFrame(
    data={
        "length": lengths,
        "token_ratio": time_token_ratio,
        "max(token_sort_ratio, token_set_ratio)": time_token_ratio_simple,
    }
)

results.to_csv("results/token_ratio.csv", sep=",", index=False)

# partial_token_ratio is unique to RapidFuzz
time_partial_token_ratio = benchmark(
    "token_ratio",
    "[rfuzz.partial_token_ratio(a, b, processor=None) for b in b_list]",
    setup,
    lengths,
    count,
)

# this gets very slow, so only benchmark it for smaller values
time_partial_token_ratio_simple = benchmark(
    "fuzzywuzzy",
    (
        "[max(rfuzz.partial_token_sort_ratio(a, b, processor=None), "
        "rfuzz.partial_token_set_ratio(a, b, processor=None)) for b in b_list]"
    ),
    setup,
    lengths,
    count,
)

results = pd.DataFrame(
    data={
        "length": lengths,
        "partial_token_ratio": time_partial_token_ratio,
        "max(partial_token_sort_ratio, partial_token_set_ratio)": time_partial_token_ratio_simple,
    }
)

results.to_csv("results/partial_token_ratio.csv", sep=",", index=False)



# SPDX-License-Identifier: MIT
# Copyright (C) 2022 Max Bachmann

from __future__ import annotations

from . import (
    OSA as OSA,
    DamerauLevenshtein as DamerauLevenshtein,
    Hamming as Hamming,
    Indel as Indel,
    Jaro as Jaro,
    JaroWinkler as JaroWinkler,
    LCSseq as LCSseq,
    Levenshtein as Levenshtein,
    Postfix as Postfix,
    Prefix as Prefix,
)
from ._initialize import (
    Editop as Editop,
    Editops as Editops,
    MatchingBlock as MatchingBlock,
    Opcode as Opcode,
    Opcodes as Opcodes,
    ScoreAlignment as ScoreAlignment,
)



# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from cpp_common cimport EditType, RfEditOp, RfOpcode, convert_string, is_valid_string
from cpython.list cimport PyList_New, PyList_SET_ITEM
from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid, PyCapsule_New
from cpython.ref cimport Py_INCREF
from libc.stdint cimport int64_t, uint32_t
from libc.stdlib cimport free, malloc
from libcpp cimport bool
from libcpp.utility cimport move
from libcpp.vector cimport vector

from rapidfuzz cimport RF_String


cdef extern from "rapidfuzz/details/types.hpp" namespace "rapidfuzz" nogil:
    cdef struct LevenshteinWeightTable:
        int64_t insert_cost
        int64_t delete_cost
        int64_t replace_cost

cdef extern from "cpp_common.hpp":
    object opcodes_apply(const RfOpcodes& ops, const RF_String& str1, const RF_String& str2) except + nogil
    object editops_apply(const RfEditops& ops, const RF_String& str1, const RF_String& str2) except + nogil

cdef str edit_type_to_str(EditType edit_type):
    if edit_type == EditType.Insert:
        return "insert"
    elif edit_type == EditType.Delete:
        return "delete"
    elif edit_type == EditType.Replace:
        return "replace"
    else:
        return "equal"

cdef EditType str_to_edit_type(edit_type) except *:
    # todo fix compiler warning about potentially uninitialized return value
    if edit_type == "insert":
        return EditType.Insert
    elif edit_type == "delete":
        return EditType.Delete
    elif edit_type == "replace":
        return EditType.Replace
    elif edit_type == "equal":
        return EditType.None
    else:
        raise ValueError("Invalid Edit Type")

cdef RfEditops list_to_editops(ops, size_t src_len, size_t dest_len) except *:
    cdef RfEditops result
    cdef size_t i
    cdef EditType edit_type
    cdef size_t src_pos, dest_pos
    cdef size_t ops_len = len(ops)
    result.set_src_len(src_len)
    result.set_dest_len(dest_len)

    if not ops_len:
        return result

    if len(ops[0]) == 5:
        return RfEditops(list_to_opcodes(ops, src_len, dest_len))

    result.reserve(ops_len)
    for op in ops:
        if len(op) != 3:
            raise TypeError("Expected list of 3-tuples, or a list of 5-tuples")

        edit_type = str_to_edit_type(op[0])
        src_pos = op[1]
        dest_pos = op[2]

        if src_pos > src_len or dest_pos > dest_len:
            raise ValueError("List of edit operations invalid")

        if src_pos == src_len and edit_type != EditType.Insert:
            raise ValueError("List of edit operations invalid")
        elif dest_pos == dest_len and edit_type != EditType.Delete:
            raise ValueError("List of edit operations invalid")

        # keep operations are not relevant in editops
        if edit_type == EditType.None:
            continue

        result.emplace_back(edit_type, src_pos, dest_pos)

    # validate order of editops
    if (result.size()):
        for i in range(0, result.size() - 1):
            if result[i + 1].src_pos < result[i].src_pos or result[i + 1].dest_pos < result[i].dest_pos:
                raise ValueError("List of edit operations out of order")
            if result[i + 1].src_pos == result[i].src_pos and result[i + 1].dest_pos == result[i].dest_pos:
                raise ValueError("Duplicated edit operation")

    result.shrink_to_fit()
    return result

cdef RfOpcodes list_to_opcodes(ops, size_t src_len, size_t dest_len) except *:
    cdef RfOpcodes result
    cdef size_t i
    cdef EditType edit_type
    cdef size_t src_start, src_end, dest_start, dest_end
    cdef size_t ops_len = len(ops)
    if not ops_len or len(ops[0]) == 3:
        return RfOpcodes(list_to_editops(ops, src_len, dest_len))

    result.set_src_len(src_len)
    result.set_dest_len(dest_len)
    result.reserve(ops_len)
    for op in ops:
        if len(op) != 5:
            raise TypeError("Expected list of 3-tuples, or a list of 5-tuples")

        edit_type = str_to_edit_type(op[0])
        src_start = op[1]
        src_end = op[2]
        dest_start = op[3]
        dest_end = op[4]

        if src_end > src_len or dest_end > dest_len:
            raise ValueError("List of edit operations invalid")
        elif src_end < src_start or dest_end < dest_start:
            raise ValueError("List of edit operations invalid")

        if edit_type == EditType.None or edit_type == EditType.Replace:
            if src_end - src_start != dest_end - dest_start or src_start == src_end:
                raise ValueError("List of edit operations invalid")
        if edit_type == EditType.Insert:
            if src_start != src_end or dest_start == dest_end:
                raise ValueError("List of edit operations invalid")
        elif edit_type == EditType.Delete:
            if src_start == src_end or dest_start != dest_end:
                raise ValueError("List of edit operations invalid")

        # merge similar adjacent blocks
        if not result.empty():
            if result.back().type == edit_type and result.back().src_end == src_start and result.back().dest_end == dest_start:
                result.back().src_end = src_end
                result.back().dest_end = dest_end
                continue

        result.emplace_back(edit_type, src_start, src_end, dest_start, dest_end)

    # check if edit operations span the complete string
    if result[0].src_begin != 0 or result[0].dest_begin != 0:
        raise ValueError("List of edit operations does not start at position 0")
    if result.back().src_end != src_len or result.back().dest_end != dest_len:
        raise ValueError("List of edit operations does not end at the string ends")
    for i in range(0, result.size() - 1):
        if result[i + 1].src_begin != result[i].src_end or result[i + 1].dest_begin != result[i].dest_end:
            raise ValueError("List of edit operations is not continuous")

    result.shrink_to_fit()
    return result

cdef list editops_to_list(const RfEditops& ops):
    cdef size_t op_count = ops.size()
    cdef list result_list = PyList_New(<Py_ssize_t>op_count)
    for i in range(op_count):
        result_item = (edit_type_to_str(ops[i].type), ops[i].src_pos, ops[i].dest_pos)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list

cdef list opcodes_to_list(const RfOpcodes& ops):
    cdef size_t op_count = ops.size()
    cdef list result_list = PyList_New(<Py_ssize_t>op_count)
    for i in range(op_count):
        result_item = (
            edit_type_to_str(ops[i].type),
            ops[i].src_begin, ops[i].src_end,
            ops[i].dest_begin, ops[i].dest_end)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list

cdef class MatchingBlock:
    """
    Triple describing matching subsequences
    """

    cdef public size_t a
    cdef public size_t b
    cdef public size_t size

    def __cinit__(self, size_t a, size_t b, size_t size):
        self.a = a
        self.b = b
        self.size = size

    def __len__(self):
        return 3

    def __eq__(self, other):
        try:
            if len(other) != 3:
                return False

            return (other[0] == self.a
                and other[1] == self.b
                and other[2] == self.size)
        except:
            return False

    def __getitem__(self, Py_ssize_t i):
        if i==0 or i==-3: return self.a
        if i==1 or i==-2: return self.b
        if i==2 or i==-1: return self.size

        raise IndexError('MatchingBlock index out of range')

    def __iter__(self):
        yield self.a
        yield self.b
        yield self.size

    def __repr__(self):
        return f"MatchingBlock(a={self.a}, b={self.b}, size={self.size})"

cdef list editops_to_matching_blocks(const RfEditops& ops):
    cdef MatchingBlock result_item
    cdef Py_ssize_t block_num = 0
    cdef size_t i = 0
    cdef size_t j = 0
    cdef size_t src_pos = 0
    cdef size_t dest_pos = 0
    for i in range(ops.size()):
        if src_pos < ops[i].src_pos or dest_pos < ops[i].dest_pos:
            if min(ops[i].src_pos - src_pos, ops[i].dest_pos - dest_pos) > 0:
                block_num += 1
            src_pos = ops[i].src_pos
            dest_pos = ops[i].dest_pos

        if ops[i].type == EditType.Replace:
            src_pos += 1
            dest_pos += 1
        elif ops[i].type == EditType.Delete:
            src_pos += 1
        elif ops[i].type == EditType.Insert:
            dest_pos += 1

    if src_pos < ops.get_src_len() or dest_pos < ops.get_dest_len():
        if min(ops.get_src_len() - src_pos, ops.get_dest_len() - dest_pos) > 0:
            block_num += 1

    cdef list result_list = PyList_New(block_num + 1)
    src_pos = 0
    dest_pos = 0
    for i in range(ops.size()):
        if src_pos < ops[i].src_pos or dest_pos < ops[i].dest_pos:
            length = min(ops[i].src_pos - src_pos, ops[i].dest_pos - dest_pos)
            if length > 0:
                result_item = MatchingBlock(src_pos, dest_pos, length)
                Py_INCREF(result_item)
                PyList_SET_ITEM(result_list, <Py_ssize_t>j, result_item)
                j += 1
            src_pos = ops[i].src_pos
            dest_pos = ops[i].dest_pos

        if ops[i].type == EditType.Replace:
            src_pos += 1
            dest_pos += 1
        elif ops[i].type == EditType.Delete:
            src_pos += 1
        elif ops[i].type == EditType.Insert:
            dest_pos += 1

    if src_pos < ops.get_src_len() or dest_pos < ops.get_dest_len():
        length = min(ops.get_src_len() - src_pos, ops.get_dest_len() - dest_pos)
        if length > 0:
            result_item = MatchingBlock(src_pos, dest_pos, length)
            Py_INCREF(result_item)
            PyList_SET_ITEM(result_list, <Py_ssize_t>j, result_item)
            j += 1

    result_item = MatchingBlock(ops.get_src_len(), ops.get_dest_len(), 0)
    Py_INCREF(result_item)
    PyList_SET_ITEM(result_list, <Py_ssize_t>block_num, result_item)
    return result_list

cdef list opcodes_to_matching_blocks(const RfOpcodes& ops):
    cdef MatchingBlock result_item
    cdef Py_ssize_t block_num = 0
    cdef size_t i = 0
    cdef size_t j = 0
    for i in range(ops.size()):
        if ops[i].type == EditType.None:
            if min(ops[i].src_end - ops[i].src_begin, ops[i].dest_end - ops[i].dest_begin) > 0:
                block_num += 1

    cdef list result_list = PyList_New(block_num + 1)
    for i in range(ops.size()):
        if ops[i].type == EditType.None:
            length = min(ops[i].src_end - ops[i].src_begin, ops[i].dest_end - ops[i].dest_begin)
            if length > 0:
                result_item = MatchingBlock(ops[i].src_begin, ops[i].dest_begin, length)
                Py_INCREF(result_item)
                PyList_SET_ITEM(result_list, <Py_ssize_t>j, result_item)
                j += 1

    result_item = MatchingBlock(ops.get_src_len(), ops.get_dest_len(), 0)
    Py_INCREF(result_item)
    PyList_SET_ITEM(result_list, <Py_ssize_t>block_num, result_item)
    return result_list

cdef class Editop:
    """
    Tuple like object describing an edit operation.
    It is in the form (tag, src_pos, dest_pos)

    The tags are strings, with these meanings:

    +-----------+---------------------------------------------------+
    | tag       | explanation                                       |
    +===========+===================================================+
    | 'replace' | src[src_pos] should be replaced by dest[dest_pos] |
    +-----------+---------------------------------------------------+
    | 'delete'  | src[src_pos] should be deleted                    |
    +-----------+---------------------------------------------------+
    | 'insert'  | dest[dest_pos] should be inserted at src[src_pos] |
    +-----------+---------------------------------------------------+
    """
    cdef public str tag
    cdef public Py_ssize_t src_pos
    cdef public Py_ssize_t dest_pos

    def __init__(self, tag, src_pos, dest_pos):
        self.tag = tag
        self.src_pos = src_pos
        self.dest_pos = dest_pos

    def __len__(self):
        return 3

    def __eq__(self, other):
        try:
            if len(other) != 3:
                return False

            return (other[0] == self.tag
                and other[1] == self.src_pos
                and other[2] == self.dest_pos)
        except:
            return False

    def __getitem__(self, Py_ssize_t i):
        if i==0 or i==-3: return self.tag
        if i==1 or i==-2: return self.src_pos
        if i==2 or i==-1: return self.dest_pos

        raise IndexError('Editop index out of range')

    def __iter__(self):
        yield self.tag
        yield self.src_pos
        yield self.dest_pos

    def __repr__(self):
        return f"Editop(tag='{self.tag}', src_pos={self.src_pos}, dest_pos={self.dest_pos})"

cdef class Editops:
    """
    List like object of Editops describing how to turn s1 into s2.
    """

    def __init__(self, editops=None, src_len=0, dest_len=0):
        if editops is not None:
            self.editops = list_to_editops(editops, src_len, dest_len)

    @classmethod
    def from_opcodes(cls, Opcodes opcodes):
        """
        Create Editops from Opcodes

        Parameters
        ----------
        opcodes : Opcodes
            opcodes to convert to editops

        Returns
        -------
        editops : Editops
            Opcodes converted to Editops
        """
        cdef Editops self = cls.__new__(cls)
        self.editops = RfEditops(opcodes.opcodes)
        return self

    def as_opcodes(self):
        """
        Convert to Opcodes

        Returns
        -------
        opcodes : Opcodes
            Editops converted to Opcodes
        """
        cdef Opcodes opcodes = Opcodes.__new__(Opcodes)
        opcodes.opcodes = RfOpcodes(self.editops)
        return opcodes

    def as_matching_blocks(self):
        """
        Convert to matching blocks

        Returns
        -------
        matching blocks : list[MatchingBlock]
            Editops converted to matching blocks
        """
        return editops_to_matching_blocks(self.editops)

    def as_list(self):
        """
        Convert Editops to a list of tuples.

        This is the equivalent of ``[x for x in editops]``
        """
        return editops_to_list(self.editops)

    def copy(self):
        """
        performs copy of Editops
        """
        cdef Editops x = Editops.__new__(Editops)
        x.editops = self.editops
        return x

    def inverse(self):
        """
        Invert Editops, so it describes how to transform the destination string to
        the source string.

        Returns
        -------
        editops : Editops
            inverted Editops

        Examples
        --------
        >>> from rapidfuzz.distance import Levenshtein
        >>> Levenshtein.editops('spam', 'park')
        [Editop(tag=delete, src_pos=0, dest_pos=0),
         Editop(tag=replace, src_pos=3, dest_pos=2),
         Editop(tag=insert, src_pos=4, dest_pos=3)]

        >>> Levenshtein.editops('spam', 'park').inverse()
        [Editop(tag=insert, src_pos=0, dest_pos=0),
         Editop(tag=replace, src_pos=2, dest_pos=3),
         Editop(tag=delete, src_pos=3, dest_pos=4)]
        """
        cdef Editops x = Editops.__new__(Editops)
        x.editops = self.editops.inverse()
        return x

    def apply(self, source_string, destination_string):
        """
        apply editops to source_string

        Parameters
        ----------
        source_string : str | bytes
            string to apply editops to
        destination_string : str | bytes
            string to use for replacements / insertions into source_string

        Returns
        -------
        mod_string : str
            modified source_string

        """
        if not (is_valid_string(source_string) and is_valid_string(destination_string)):
            raise TypeError("expected strings or bytes object")

        return editops_apply(self.editops, convert_string(source_string), convert_string(destination_string))

    def remove_subsequence(self, Editops subsequence):
        """
        remove a subsequence

        Parameters
        ----------
        subsequence : Editops
            subsequence to remove (has to be a subset of editops)

        Returns
        -------
        sequence : Editops
            a copy of the editops without the subsequence
        """
        cdef Editops x = Editops.__new__(Editops)
        x.editops = self.editops.remove_subsequence(subsequence.editops)
        return x

    @property
    def src_len(self):
        return self.editops.get_src_len()

    @src_len.setter
    def src_len(self, value):
        self.editops.set_src_len(value)

    @property
    def dest_len(self):
        return self.editops.get_dest_len()

    @dest_len.setter
    def dest_len(self, value):
        self.editops.set_dest_len(value)

    def __eq__(self, other):
        if isinstance(other, Editops):
            return self.editops == (<Editops>other).editops

        return False

    def __len__(self):
        return self.editops.size()

    def __delitem__(self, key):
        cdef Py_ssize_t index
        cdef Py_ssize_t start, stop, step

        if isinstance(key, int):
            index = key
            if index < 0:
                index += <Py_ssize_t>self.editops.size()

            if index < 0 or index >= <Py_ssize_t>self.editops.size():
                raise IndexError("Editops index out of range")

            self.editops.erase(self.editops.begin() + index)
        elif isinstance(key, slice):
            start, stop, step = key.indices(<Py_ssize_t>self.editops.size())
            if step < 0:
                raise ValueError("step sizes below 0 lead to an invalid order of editops")

            self.editops.remove_slice(start, stop, step)
        else:
            raise TypeError("Expected index or slice")

    def __getitem__(self, key):
        cdef Py_ssize_t index
        cdef Py_ssize_t start, stop, step

        if isinstance(key, int):
            index = key
            if index < 0:
                index += <Py_ssize_t>self.editops.size()

            if index < 0 or index >= <Py_ssize_t>self.editops.size():
                raise IndexError("Editops index out of range")

            return Editop(
                edit_type_to_str(self.editops[index].type),
                self.editops[index].src_pos,
                self.editops[index].dest_pos
            )
        elif isinstance(key, slice):
            start, stop, step = key.indices(<Py_ssize_t>self.editops.size())
            if step < 0:
                raise ValueError("step sizes below 0 lead to an invalid order of editops")
            x = Editops.__new__(Editops)
            (<Editops>x).editops = self.editops.slice(start, stop, step)
            return x
        else:
            raise TypeError("Expected index or slice")

    def __iter__(self):
        cdef size_t i
        for i in range(self.editops.size()):
            yield Editop(
                edit_type_to_str(self.editops[i].type),
                self.editops[i].src_pos,
                self.editops[i].dest_pos
            )

    def __repr__(self):
        return "Editops([" + ", ".join(repr(op) for op in self) + f"], src_len={self.editops.get_src_len()}, dest_len={self.editops.get_dest_len()})"

cdef class Opcode:
    """
    Tuple like object describing an edit operation.
    It is in the form (tag, src_start, src_end, dest_start, dest_end)

    The tags are strings, with these meanings:

    +-----------+-----------------------------------------------------+
    | tag       | explanation                                         |
    +===========+=====================================================+
    | 'replace' | src[src_start:src_end] should be                    |
    |           | replaced by dest[dest_start:dest_end]               |
    +-----------+-----------------------------------------------------+
    | 'delete'  | src[src_start:src_end] should be deleted.           |
    |           | Note that dest_start==dest_end in this case.        |
    +-----------+-----------------------------------------------------+
    | 'insert'  | dest[dest_start:dest_end] should be inserted        |
    |           | at src[src_start:src_start].                        |
    |           | Note that src_start==src_end in this case.          |
    +-----------+-----------------------------------------------------+
    | 'equal'   | src[src_start:src_end] == dest[dest_start:dest_end] |
    +-----------+-----------------------------------------------------+

    Note
    ----
    Opcode is compatible with the tuples returned by difflib's SequenceMatcher to make them
    interoperable
    """
    cdef public str tag
    cdef public Py_ssize_t src_start
    cdef public Py_ssize_t src_end
    cdef public Py_ssize_t dest_start
    cdef public Py_ssize_t dest_end

    def __init__(self, tag, src_start, src_end, dest_start, dest_end):
        self.tag = tag
        self.src_start = src_start
        self.src_end = src_end
        self.dest_start = dest_start
        self.dest_end = dest_end

    def __len__(self):
        return 5

    def __eq__(self, other):
        try:
            if len(other) != 5:
                return False

            return (other[0] == self.tag
                and other[1] == self.src_start
                and other[2] == self.src_end
                and other[3] == self.dest_start
                and other[4] == self.dest_end)
        except:
            return False

    def __getitem__(self, Py_ssize_t i):
        if i==0 or i==-5: return self.tag
        if i==1 or i==-4: return self.src_start
        if i==2 or i==-3: return self.src_end
        if i==3 or i==-2: return self.dest_start
        if i==4 or i==-1: return self.dest_end

        raise IndexError('Opcode index out of range')

    def __iter__(self):
        yield self.tag
        yield self.src_start
        yield self.src_end
        yield self.dest_start
        yield self.dest_end

    def __repr__(self):
        return f"Opcode(tag='{self.tag}', src_start={self.src_start}, src_end={self.src_end}, dest_start={self.dest_start}, dest_end={self.dest_end})"

cdef class Opcodes:
    """
    List like object of Opcodes describing how to turn s1 into s2.
    The first Opcode has src_start == dest_start == 0, and remaining tuples
    have src_start == the src_end from the tuple preceding it,
    and likewise for dest_start == the previous dest_end.
    """

    def __init__(self, opcodes=None, src_len=0, dest_len=0):
        if opcodes is not None:
            self.opcodes = list_to_opcodes(opcodes, src_len, dest_len)

    @classmethod
    def from_editops(cls, Editops editops):
        """
        Create Opcodes from Editops

        Parameters
        ----------
        editops : Editops
            editops to convert to opcodes

        Returns
        -------
        opcodes : Opcodes
            Editops converted to Opcodes
        """
        cdef Opcodes self = cls.__new__(cls)
        self.opcodes = RfOpcodes(editops.editops)
        return self

    def as_editops(self):
        """
        Convert to Editops

        Returns
        -------
        editops : Editops
            Opcodes converted to Editops
        """
        cdef Editops editops = Editops.__new__(Editops)
        editops.editops = RfEditops(self.opcodes)
        return editops

    def as_matching_blocks(self):
        """
        Convert to matching blocks

        Returns
        -------
        matching blocks : list[MatchingBlock]
            Opcodes converted to matching blocks
        """
        return opcodes_to_matching_blocks(self.opcodes)

    def as_list(self):
        """
        Convert Opcodes to a list of tuples, which is compatible
        with the opcodes of difflibs SequenceMatcher.

        This is the equivalent of ``[x for x in opcodes]``
        """
        return opcodes_to_list(self.opcodes)

    def copy(self):
        """
        performs copy of Opcodes
        """
        cdef Opcodes x = Opcodes.__new__(Opcodes)
        x.opcodes = self.opcodes
        return x

    def inverse(self):
        """
        Invert Opcodes, so it describes how to transform the destination string to
        the source string.

        Returns
        -------
        opcodes : Opcodes
            inverted Opcodes

        Examples
        --------
        >>> from rapidfuzz.distance import Levenshtein
        >>> Levenshtein.opcodes('spam', 'park')
        [Opcode(tag=delete, src_start=0, src_end=1, dest_start=0, dest_end=0),
         Opcode(tag=equal, src_start=1, src_end=3, dest_start=0, dest_end=2),
         Opcode(tag=replace, src_start=3, src_end=4, dest_start=2, dest_end=3),
         Opcode(tag=insert, src_start=4, src_end=4, dest_start=3, dest_end=4)]

        >>> Levenshtein.opcodes('spam', 'park').inverse()
        [Opcode(tag=insert, src_start=0, src_end=0, dest_start=0, dest_end=1),
         Opcode(tag=equal, src_start=0, src_end=2, dest_start=1, dest_end=3),
         Opcode(tag=replace, src_start=2, src_end=3, dest_start=3, dest_end=4),
         Opcode(tag=delete, src_start=3, src_end=4, dest_start=4, dest_end=4)]
        """
        cdef Opcodes x = Opcodes.__new__(Opcodes)
        x.opcodes = self.opcodes.inverse()
        return x

    def apply(self, source_string, destination_string):
        """
        apply opcodes to source_string

        Parameters
        ----------
        source_string : str | bytes
            string to apply opcodes to
        destination_string : str | bytes
            string to use for replacements / insertions into source_string

        Returns
        -------
        mod_string : str
            modified source_string

        """
        if not (is_valid_string(source_string) and is_valid_string(destination_string)):
            raise TypeError("expected strings or bytes object")

        return opcodes_apply(self.opcodes, convert_string(source_string), convert_string(destination_string))

    @property
    def src_len(self):
        return self.opcodes.get_src_len()

    @src_len.setter
    def src_len(self, value):
        self.opcodes.set_src_len(value)

    @property
    def dest_len(self):
        return self.opcodes.get_dest_len()

    @dest_len.setter
    def dest_len(self, value):
        self.opcodes.set_dest_len(value)

    def __eq__(self, other):
        if isinstance(other, Opcodes):
            return self.opcodes == (<Opcodes>other).opcodes

        return False

    def __len__(self):
        return self.opcodes.size()

    def __getitem__(self, key):
        cdef Py_ssize_t index

        if isinstance(key, int):
            index = key
            if index < 0:
                index += <Py_ssize_t>self.opcodes.size()

            if index < 0 or index >= <Py_ssize_t>self.opcodes.size():
                raise IndexError("Opcodes index out of range")

            return Opcode(
                edit_type_to_str(self.opcodes[index].type),
                self.opcodes[index].src_begin,
                self.opcodes[index].src_end,
                self.opcodes[index].dest_begin,
                self.opcodes[index].dest_end
            )
        else:
            raise TypeError("Expected index")

    def __iter__(self):
        cdef size_t i
        for i in range(self.opcodes.size()):
            yield Opcode(
                edit_type_to_str(self.opcodes[i].type),
                self.opcodes[i].src_begin,
                self.opcodes[i].src_end,
                self.opcodes[i].dest_begin,
                self.opcodes[i].dest_end
            )

    def __repr__(self):
        return "Opcodes([" + ", ".join(repr(op) for op in self) + f"], src_len={self.opcodes.get_src_len()}, dest_len={self.opcodes.get_dest_len()})"


cdef class ScoreAlignment:
    """
    Tuple like object describing the position of the compared strings in
    src and dest.

    It indicates that the score has been calculated between
    src[src_start:src_end] and dest[dest_start:dest_end]
    """
    def __init__(self, score, src_start, src_end, dest_start, dest_end):
        self.score = score
        self.src_start = src_start
        self.src_end = src_end
        self.dest_start = dest_start
        self.dest_end = dest_end

    def __len__(self):
        return 5

    def __eq__(self, other):
        try:
            if len(other) != 5:
                return False

            return (other[0] == self.score
                and other[1] == self.src_start
                and other[2] == self.src_end
                and other[3] == self.dest_start
                and other[4] == self.dest_end)
        except:
            return False

    def __getitem__(self, Py_ssize_t i):
        if i==0 or i==-5: return self.score
        if i==1 or i==-4: return self.src_start
        if i==2 or i==-3: return self.src_end
        if i==3 or i==-2: return self.dest_start
        if i==4 or i==-1: return self.dest_end

        raise IndexError('Opcode index out of range')

    def __iter__(self):
        yield self.score
        yield self.src_start
        yield self.src_end
        yield self.dest_start
        yield self.dest_end

    def __repr__(self):
        return f"ScoreAlignment(score={self.score}, src_start={self.src_start}, src_end={self.src_end}, dest_start={self.dest_start}, dest_end={self.dest_end})"



from __future__ import annotations

import timeit

import numpy as np
import pandas as pd


def benchmark(name, func, setup, lengths, count):
    print(f"starting {name}")
    start = timeit.default_timer()
    results = []
    for length in lengths:
        test = timeit.Timer(func, setup=setup.format(length, count))
        results.append(min(test.timeit(number=1) for _ in range(7)) / count)
    stop = timeit.default_timer()
    print(f"finished {name}, Runtime: ", stop - start)
    return results


setup = """
from rapidfuzz import fuzz as rfuzz
from fuzzywuzzy import fuzz
import string
import random
random.seed(18)
characters = string.ascii_letters + string.digits + string.whitespace + string.punctuation
a      = ''.join(random.choice(characters) for _ in range(64))
b_list = [''.join(random.choice(characters) for _ in range({0})) for _ in range({1})]
"""

lengths = list(range(64, 512, 2))
count = 4000

time_rapidfuzz = benchmark("rapidfuzz", "[rfuzz.partial_ratio(a, b) for b in b_list]", setup, lengths, count)

# this gets very slow, so only benchmark it for smaller values
time_fuzzywuzzy = (
    benchmark(
        "fuzzywuzzy",
        "[fuzz.partial_ratio(a, b) for b in b_list]",
        setup,
        list(range(64, 256, 2)),
        count,
    )
    + [np.nan] * 128
)

results = pd.DataFrame(
    data={
        "length": lengths,
        "rapidfuzz": time_rapidfuzz,
        "fuzzywuzzy": time_fuzzywuzzy,
    }
)

results.to_csv("results/partial_ratio_short_needle.csv", sep=",", index=False)



"""
rapid string matching library
"""

from __future__ import annotations

__author__: str = "Max Bachmann"
__license__: str = "MIT"
__version__: str = "3.12.2"

from rapidfuzz import distance, fuzz, process, utils

__all__ = ["distance", "fuzz", "process", "utils", "get_include"]


def get_include():
    """
    Return the directory that contains the RapidFuzz \\*.h header files.
    Extension modules that need to compile against RapidFuzz should use this
    function to locate the appropriate include directory.
    Notes
    -----
    When using ``distutils``, for example in ``setup.py``.
    ::
        import rapidfuzz_capi
        ...
        Extension('extension_name', ...
                include_dirs=[rapidfuzz_capi.get_include()])
        ...
    """
    from pathlib import Path

    return str(Path(__file__).parent)
